from .base import BaseQuery
from .from_row import FromRowQuery
from .join_query import JoinQuery
from .pagination_query import PaginationQuery
from .sorting_query import SortingQuery


class SelectQuery(
    FromRowQuery,
    JoinQuery,
    PaginationQuery,
    SortingQuery,
    BaseQuery
):
    pass



from typing import Union, Type
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.sql import Select, Insert, Update, Delete
from query.join_query import JoinQuery
from .base import BaseProvider


class JoinProvider(BaseProvider):
    def _join(
        self,
        field_name: str,
        query: Union[JoinQuery, Type[JoinQuery]],
        stmt: Union[Select, Insert, Update, Delete],
        mapper: DeclarativeMeta
    ) -> Union[Select, Insert, Update, Delete]:
        mapper_field = getattr(mapper, field_name)
        join_strategy = query.get_join_strategy(field_name=field_name)

        return stmt.join(
            mapper_field,
            isouter=join_strategy.is_outer,
            full=join_strategy.is_full
        )



from abc import abstractmethod
from typing import Any, Dict, Sequence, Optional
from sqlalchemy.orm import DeclarativeMeta, ColumnProperty
from sqlalchemy.sql import update, Update
from query.update_query import UpdateQuery
from .base import BaseProvider
from .join_provider import JoinProvider


class UpdateProvider(JoinProvider, BaseProvider):
    @abstractmethod
    async def update(self, *args, **kwargs):
        pass

    def make_update_stmt(
        self,
        query: UpdateQuery,
        mapper: DeclarativeMeta,
        returning: bool = True
    ) -> Update:
        update_stmt = update(mapper)
        update_stmt = self._bind_clause(
            clause=query.get_filters(),
            mapper=mapper,
            stmt=update_stmt,
        )

        updatable_values = self.__make_updatable_values(
            query=query,
            mapper=mapper,
        )
        if not updatable_values:
            raise ValueError(
                'Attr values not empty'
            )

        update_stmt = update_stmt.values(**updatable_values)

        if returning:
            update_stmt = update_stmt.returning(mapper)

        return update_stmt

    def make_bulk_update_stmt(self):
        """
        Will be released in next version
        Use simple make_update_stmt for simple bulk_updates
        """
        raise NotImplementedError

    async def _update(
        self,
        query: UpdateQuery,
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Optional[Sequence[UpdateQuery]]:
        """
        Returns sequence of query instance if returning is True
        """
        update_stmt = self.make_update_stmt(
            query=query,
            mapper=mapper,
            returning=returning
        )

        # import pdb
        # pdb.set_trace()

        scalar_result = await self._session.execute(update_stmt)

        if not returning:
            return

        return query.from_returning_mappers(scalar_result.all())

    def __make_updatable_values(
        self,
        query: UpdateQuery,
        mapper: DeclarativeMeta
    ) -> Dict[str, Any]:
        updatable_values = dict()

        for field_name, value in query.get_values().items():
            mapper_field = getattr(mapper, field_name, None)
            if mapper_field is None:
                continue

            if not isinstance(mapper_field.property, ColumnProperty):
                continue

            updatable_values[field_name] = value

        return updatable_values



from __future__ import annotations
from abc import ABC
from .base import BaseQuery
from .select_query import SelectQuery
from .insert_query import InsertQuery
from .update_query import UpdateQuery
from .delete_query import DeleteQuery


class AbstractQuery(
    ABC,
    SelectQuery,
    InsertQuery,
    UpdateQuery,
    DeleteQuery,
    BaseQuery
):
    pass



from .base import BaseQuery


class JoinStrategy:
    is_outer: bool = False
    is_full: bool = False

    def __init__(self, is_outer: bool = False, is_full: bool = False):
        self.is_outer = is_outer
        self.is_full = is_full


class JoinQuery(BaseQuery):
    @classmethod
    def get_join_strategy(cls, field_name: str):
        class_ = cls
        if isinstance(cls, BaseQuery):
            class_ = cls.__class__

        return getattr(class_, field_name, JoinStrategy())



from types import MappingProxyType
from typing import Any, Mapping, Callable, Optional, Union
from sqlalchemy.orm import DeclarativeMeta, InstrumentedAttribute
from sqlalchemy.sql import Select, Insert, Update, Delete
from sqlalchemy.sql.expression import BinaryExpression
from utils import is_relationship, get_column, get_related_mapper, is_column
from .base import BaseClauseBinder


class StringClauseBuilder(BaseClauseBinder):
    EQUAL_OPERATOR = 'e'
    NOT_EQUAL_OPERATOR = 'ne'
    LESS_THAN_OPERATOR = 'l'
    LESS_THAN_OR_EQUAL_TO_OPERATOR = 'le'
    GREATER_THAN_OPERATOR = 'g'
    GREATER_THAN_OR_EQUAL_TO_OPERATOR = 'ge'
    LIKE_OPERATOR = 'like'
    ILIKE_OPERATOR = 'ilike'
    IN_OPERATOR = 'in'
    NOT_IN_OPERATOR = 'not_in'

    LOOKUP_OPERATORS: Mapping[
        str,
        Callable[[InstrumentedAttribute, Any], BinaryExpression]
    ] = MappingProxyType({
        EQUAL_OPERATOR:
            lambda _column, _value: _column == _value,
        NOT_EQUAL_OPERATOR:
            lambda _column, _value: _column != _value,
        LESS_THAN_OPERATOR:
            lambda _column, _value: _column < _value,
        LESS_THAN_OR_EQUAL_TO_OPERATOR:
            lambda _column, _value: _column <= _value,
        GREATER_THAN_OPERATOR:
            lambda _column, _value: _column > _value,
        GREATER_THAN_OR_EQUAL_TO_OPERATOR:
            lambda _column, _value: _column >= _value,
        LIKE_OPERATOR:
            lambda _column, _value: _column.like(_value),
        ILIKE_OPERATOR:
            lambda _column, _value: _column.ilike(_value),
        IN_OPERATOR:
            lambda _column, _value: _column.in_(_value),
        NOT_IN_OPERATOR:
            lambda _column, _value: _column.not_in(_value),
    })

    def _get_column(
        self,
        lookup: str,
        mapper: DeclarativeMeta
    ) -> Optional[InstrumentedAttribute]:
        """
        lookup: meter_inline__directory__name__ilike: "%some_directory_name%"
        """
        lookup_parts = lookup.split(self.LOOKUP_STRING)
        for i, lookup_part in enumerate(lookup_parts):
            mapper_field = get_column(mapper=mapper, field_name=lookup_part)
            if mapper_field is None:
                return

            if is_column(mapper_field=mapper_field):
                return mapper_field

            if is_relationship(mapper_field=mapper_field):
                return self._get_column(
                    lookup=self.LOOKUP_STRING.join(lookup_parts[i+1:]),
                    mapper=get_related_mapper(
                        mapper=mapper,
                        field_name=lookup_part
                    )
                )

    def _get_expression(
        self,
        lookup: str,
        value: Any,
        mapper: DeclarativeMeta,
    ) -> Optional[BinaryExpression]:
        """
        """
        column = self._get_column(
            lookup=lookup,
            mapper=mapper
        )
        if column is None:
            return

        operator_names = lookup.split(self.LOOKUP_STRING)
        if not operator_names:
            return

        operator_name = operator_names[-1]
        if operator_name not in self.LOOKUP_OPERATORS:
            operator_name = self.EQUAL_OPERATOR

        lookup_operator = self.LOOKUP_OPERATORS.get(operator_name)

        if lookup_operator is None:
            return

        return lookup_operator(column, value)

    def _bind_string_expression_method(
        self,
        lookup: str,
        value: Any,
        mapper: DeclarativeMeta,
        stmt: Union[Select, Insert, Update, Delete],
    ) -> Select:
        expression = self._get_expression(
            lookup=lookup,
            value=value,
            mapper=mapper,
        )
        if expression is None:
            return stmt

        return stmt.where(expression)



from abc import abstractmethod
from typing import Any, Optional, List, Mapping, Sequence, Type, Union, Dict
from sqlalchemy.orm import DeclarativeMeta, ColumnProperty
from sqlalchemy.sql import insert, Insert
from query.insert_query import InsertQuery
from .base import BaseProvider


class InsertProvider(BaseProvider):
    @abstractmethod
    async def insert(self, *args, **kwargs):
        pass

    @abstractmethod
    async def bulk_insert(self, *args, **kwargs):
        pass

    def make_insert_stmt(
        self,
        query: InsertQuery,
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Insert:
        insert_stmt = insert(mapper)

        insertable_values = self.__make_insertable_values(
            query=query,
            mapper=mapper,
            values=query.dict,
        )

        insert_stmt = insert_stmt.values(**insertable_values)

        if returning:
            insert_stmt = insert_stmt.returning(mapper)

        return insert_stmt

    def make_bulk_insert_stmt(
        self,
        query: Union[InsertQuery, Type[InsertQuery]],
        values_seq: Sequence[Dict[str, Any]],
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Insert:
        """
        """
        insertable_values_seq: List[Mapping[str, Any]] = []

        for values in values_seq:
            insertable_values = self.__make_insertable_values(
                query=query,
                mapper=mapper,
                values=values
            )
            insertable_values_seq.append(insertable_values)

        insert_stmt = insert(mapper)
        insert_stmt = insert_stmt.values(insertable_values_seq)

        if returning:
            insert_stmt = insert_stmt.returning(mapper)

        return insert_stmt

    async def _insert(
        self,
        query: InsertQuery,
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Optional[InsertQuery]:
        """
        if returning is True, returns instance of passed query
        """
        insert_stmt = self.make_insert_stmt(
            query=query,
            mapper=mapper,
            returning=returning
        )

        scalar_result = await self._session.execute(insert_stmt)

        if returning:
            return query.from_returning_mapper(scalar_result.first())

    async def _bulk_insert(
        self,
        query: Union[InsertQuery, Type[InsertQuery]],
        values_seq: Sequence[Dict[str, Any]],
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Optional[Sequence[InsertQuery]]:
        """
        If returning is True, returns iterable object that contains
        InsertQuery
        """
        if not values_seq:
            return

        insert_stmt = self.make_bulk_insert_stmt(
            query=query,
            values_seq=values_seq,
            mapper=mapper,
            returning=returning
        )

        scalar_result = await self._session.execute(insert_stmt)

        if returning:
            return query.from_returning_mappers(scalar_result.all())

    def __make_insertable_values(
        self,
        query: InsertQuery,
        mapper: DeclarativeMeta,
        values: Dict[str, Any],
    ) -> Mapping[str, Any]:
        insertable_values = dict()

        for field_name, value in values.items():
            if field_name not in query.get_type_hints():
                continue

            mapper_field = getattr(mapper, field_name, None)

            if mapper_field is None:
                continue

            if not isinstance(mapper_field.property, ColumnProperty):
                continue

            insertable_values[field_name] = value

        return insertable_values



from abc import abstractmethod
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.sql import delete, Delete
from query.delete_query import DeleteQuery
from .base import BaseProvider
from .join_provider import JoinProvider


class DeleteProvider(JoinProvider, BaseProvider):
    @abstractmethod
    async def delete(self, *args, **kwargs):
        pass

    def make_delete_stmt(
        self,
        query: DeleteQuery,
        mapper: DeclarativeMeta,
    ) -> Delete:
        delete_stmt = delete(mapper)
        delete_stmt = self._bind_clause(
            clause=query.get_filters(),
            mapper=mapper,
            stmt=delete_stmt
        )
        return delete_stmt

    async def _delete(
        self,
        query: DeleteQuery,
        mapper: DeclarativeMeta
    ):
        delete_stmt = self.make_delete_stmt(
            query=query,
            mapper=mapper
        )
        await self._session.execute(delete_stmt)



from typing import List, Dict, Any, Union, Optional, Sequence
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.sql import Select, Insert, Update, Delete
from clause_binder.clause_binder import ClauseBinder
from query.query import AbstractQuery
from .select_provider import SelectProvider
from .insert_provider import InsertProvider
from .update_provider import UpdateProvider
from .delete_provider import DeleteProvider


class Provider(
    SelectProvider,
    InsertProvider,
    UpdateProvider,
    DeleteProvider
):
    def _bind_clause(
        self,
        clause: Dict[str, Any],
        mapper: DeclarativeMeta,
        stmt: Union[Select, Insert, Update, Delete],
    ) -> Union[Select, Insert, Update, Delete]:
        stmt = ClauseBinder().bind(
            clause=clause,
            mapper=mapper,
            stmt=stmt,
        )
        return stmt

    async def select(
        self,
        query: AbstractQuery,
        mapper: DeclarativeMeta
    ) -> List[AbstractQuery]:
        return await self._select(
            query=query,
            mapper=mapper
        )

    async def insert(
        self,
        query: AbstractQuery,
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Optional[AbstractQuery]:
        return await self._insert(
            query=query,
            mapper=mapper,
            returning=returning
        )

    async def bulk_insert(
        self,
        queries: Sequence[AbstractQuery],
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Optional[Sequence[AbstractQuery]]:
        return await self._bulk_insert(
            queries=queries,
            mapper=mapper,
            returning=returning
        )

    async def update(
        self,
        query: AbstractQuery,
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Optional[Sequence[AbstractQuery]]:
        return await self._update(
            query=query,
            mapper=mapper,
            returning=returning
        )

    async def delete(
        self,
        query: AbstractQuery,
        mapper: DeclarativeMeta
    ):
        await self._delete(
            query=query,
            mapper=mapper
        )



from __future__ import annotations
from copy import deepcopy
from typing import Any, get_type_hints, Type, get_args, Tuple, Dict
from utils import cls_or_ins
from .field import Field


class BaseQuery:
    _filters: Dict[str, Any] = dict()

    def __init__(self, **kwargs):
        self._set_attrs(**kwargs)

    @staticmethod
    def _set_values(mapping: Dict[str, Any], **kwargs):
        for field, value in kwargs.items():
            if value is ...:
                continue

            mapping[field] = value

    @staticmethod
    def _get_query_type(type_hint: Any) -> Type[BaseQuery]:
        try:
            if issubclass(type_hint, BaseQuery):
                return type_hint
        except:
            pass

        args = get_args(type_hint)
        answers = []
        for item in args:
            answers.append(BaseQuery._get_query_type(item))

        for answer in answers:
            try:
                if issubclass(answer, BaseQuery):
                    return answer
            except:
                pass

    @cls_or_ins
    def is_class(cls_or_ins) -> bool:
        return not isinstance(cls_or_ins, BaseQuery) \
               and issubclass(cls_or_ins, BaseQuery)

    @cls_or_ins
    def is_instance(cls_or_ins) -> bool:
        return isinstance(cls_or_ins, BaseQuery)

    @cls_or_ins
    def get_filters(cls_or_ins) -> Dict[str, Any]:
        return deepcopy(cls_or_ins._filters)

    @cls_or_ins
    def set_filters(cls_or_ins, **kwargs) -> BaseQuery:
        self = cls_or_ins
        if cls_or_ins.is_class():
            self = cls_or_ins()

        self._set_values(self._filters, **kwargs)

        return self

    @cls_or_ins
    def get_class(cls_or_ins) -> Type[BaseQuery]:
        if cls_or_ins.is_instance:
            return cls_or_ins.__class__

        return cls_or_ins

    @classmethod
    def get_name(cls) -> str:
        return cls.__name__

    @classmethod
    def get_type_hints(cls):
        class_ = cls
        if isinstance(cls, BaseQuery):
            class_ = cls.__class__

        return get_type_hints(class_)

    @classmethod
    def get_field_query(cls, field_name: str) -> Type[BaseQuery]:
        field_type_hint = cls.get_type_hints().get(field_name)
        if field_type_hint is None:
            raise AttributeError

        field = cls._get_query_type(field_type_hint)

        if field is None or not issubclass(field, BaseQuery):
            raise KeyError(
                f'Attribute {field_name}, '
                f'in {cls.__name__ or cls.__class__.__name__} is not Query type'
            )

        return field

    @classmethod
    def get_fields_count(cls) -> int:
        return len(cls.get_type_hints())

    @classmethod
    def _get_field_type(cls, field_name: str) -> Type:
        type_hints = cls.get_type_hints()
        if field_name in type_hints:
            return type_hints[field_name]

        return type(getattr(cls, field_name, Any))

    @classmethod
    def _get_field_types(cls, field_name: str) -> Tuple[Type]:
        field_type = cls._get_field_type(field_name=field_name)
        return getattr(field_type, '__args__', (field_type, ))

    @classmethod
    def _is_query_field(cls, field_name: str) -> bool:
        field_types = cls._get_field_types(field_name=field_name)
        for type_ in field_types:
            if issubclass(type_, BaseQuery):
                return True

        return False

    @property
    def dict(self) -> Dict[str, Any]:
        mapping = dict()
        for field, value in self.__dict__.items():
            if field.startswith('_'):
                continue

            if isinstance(value, BaseQuery):
                mapping[field] = value.dict

            mapping[field] = value

        return mapping

    def _set_attrs(self, **kwargs):
        for field, value in kwargs.items():
            if value is ...:
                value = None

            if field not in self.get_type_hints():
                continue

            attr_default = getattr(self, field, None)
            if isinstance(attr_default, Field):
                setattr(self, field, attr_default(value))
                continue

            setattr(self, field, value)



from typing import Any, Union, Dict
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.sql import Select, Insert, Update, Delete
from .self_method_binder import SelfMethodBinder
from .string_clause_binder import StringClauseBuilder


class ClauseBinder(SelfMethodBinder, StringClauseBuilder):
    def bind(
        self,
        clause: Dict[str, Any],
        mapper: DeclarativeMeta,
        stmt: Union[Select, Insert, Update, Delete]
    ) -> Select:
        return self._bind(
            clause=clause,
            mapper=mapper,
            stmt=stmt,
        )



from abc import ABC
from typing import List, Dict, Any, Union, Optional, Sequence, Type
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.sql import Select, Insert, Update, Delete
from clause_binder.abstract_clause_binder import AbstractClauseBinder
from query.base import BaseQuery
from query.query import AbstractQuery
from .select_provider import SelectProvider
from .insert_provider import InsertProvider
from .update_provider import UpdateProvider
from .delete_provider import DeleteProvider


class AbstractQueryProvider(
    ABC,
    SelectProvider,
    InsertProvider,
    UpdateProvider,
    DeleteProvider
):
    _mapper: DeclarativeMeta
    _clause_binder: AbstractClauseBinder
    _query_type: Type[AbstractQuery]

    def make_select_stmt(
        self,
        **kwargs
    ) -> Select:
        return super().make_select_stmt(
            query=self._query_type.set_filters(**kwargs),
            mapper=self._mapper
        )

    async def select(
        self,
        **kwargs
    ) -> List[BaseQuery]:
        return await self._select(
            query=self._query_type.set_filters(**kwargs),
            mapper=self._mapper
        )

    async def insert(
        self,
        returning: bool = True,
        **kwargs
    ) -> Optional[AbstractQuery]:
        return await self._insert(
            query=self._query_type(**kwargs),
            mapper=self._mapper,
            returning=returning
        )

    async def bulk_insert(
        self,
        values_seq: Sequence[Dict[str, Any]],
        returning: bool = True,
    ) -> Optional[Sequence[AbstractQuery]]:
        return await self._bulk_insert(
            query=self._query_type,
            values_seq=values_seq,
            mapper=self._mapper,
            returning=returning
        )

    async def update(
        self,
        filters: Dict[str, Any],
        values: Dict[str, Any],
        returning: bool = True,
    ) -> Optional[Sequence[AbstractQuery]]:
        return await self._update(
            query=self._query_type().set_filters(**filters).set_values(**values),
            mapper=self._mapper,
            returning=returning
        )

    async def delete(
        self,
        **kwargs
    ):
        await self._delete(
            query=self._query_type.set_filters(**kwargs),
            mapper=self._mapper
        )

    def _bind_clause(
        self,
        clause: Dict[str, Any],
        stmt: Union[Select, Insert, Update, Delete],
        *args, **kwargs
    ) -> Union[Select, Insert, Update, Delete]:
        stmt = self._clause_binder.bind(
            clause=clause,
            stmt=stmt,
        )

        return stmt



from typing import Any, Callable, Optional


class Field:
    _adapter: Optional[Callable[[Any], Any]] = None

    def __init__(
        self,
        adapter: Optional[Callable[[Any], Any]] = None,
    ):
        self._adapter = adapter

    def __call__(self, value: Any) -> Any:
        if self._adapter is not None:
            return self._adapter(value)

        return value



from __future__ import annotations
from abc import abstractmethod
from typing import Union, Type, List, Optional
from sqlalchemy import select
from sqlalchemy.sql import Select, Insert
from sqlalchemy.orm import DeclarativeMeta, ColumnProperty, \
    RelationshipProperty, InstrumentedAttribute
from query.base import BaseQuery
from query.from_row import FIELD_NAME_SEPARATOR
from query.select_query import SelectQuery
from utils import get_related_mapper
from .base import BaseProvider
from .join_provider import JoinProvider
from .pagination_provider import PaginationProvider
from .sorting_provider import SortingProvider


class SelectProvider(
    JoinProvider,
    PaginationProvider,
    SortingProvider,
    BaseProvider
):
    @abstractmethod
    async def select(self, *args, **kwargs):
        pass

    def make_select_stmt(
        self,
        query: Union[Type[SelectQuery], SelectQuery],
        mapper: DeclarativeMeta
    ) -> Select:
        select_stmt = select()

        select_stmt = self._make_select_stmt(
            select_stmt=select_stmt,
            query=query,
            mapper=mapper
        )

        select_stmt = self.bind_pagination(
            query=query,
            select_stmt=select_stmt
        )

        select_stmt = self.bind_sorting(
            query=query,
            mapper=mapper,
            select_stmt=select_stmt,
        )

        if query.is_instance():
            select_stmt = self._bind_clause(
                clause=query.get_filters(),
                mapper=mapper,
                stmt=select_stmt
            )

        return select_stmt

    def make_select_from_insert(
        self,
        query: SelectQuery,
        insert_stmt: Insert
    ) -> Select:
        """
        Will be added in future version

        with inserted as (
            insert intoROW_MAP_FORMAT.format(
                        query.get_name(), field_name
                    ) test2(name, description, test_id) values
            ('test2_name1', 'test2_description1', 1)
            returning *
        )
        select inserted.*, test.*
        from inserted left outer join test on inserted.test_id = test.id
        """
        raise NotImplementedError

    def _make_column_label(
        self,
        mapper_field: InstrumentedAttribute,
        label_prefix: Optional[str] = None,
    ) -> str:
        if label_prefix is None:
            return mapper_field.name

        return label_prefix + FIELD_NAME_SEPARATOR + mapper_field.name

    def _make_select_stmt(
        self,
        select_stmt: Select,
        query: Union[Type[SelectQuery], SelectQuery],
        mapper: DeclarativeMeta,
        label_prefix: Optional[str] = None,
    ) -> Select:
        type_hints = query.get_type_hints()

        for field_name, type_hint in type_hints.items():
            mapper_field = getattr(mapper, field_name, None)
            if mapper_field is None:
                continue

            if isinstance(mapper_field.property, ColumnProperty):
                select_stmt = select_stmt.add_columns(
                    mapper_field.label(
                        self._make_column_label(
                            mapper_field=mapper_field,
                            label_prefix=label_prefix
                        )
                    )
                )
                continue

            if isinstance(mapper_field.property, RelationshipProperty):
                select_stmt = self._join(
                    field_name=field_name,
                    query=query,
                    stmt=select_stmt,
                    mapper=mapper,
                )
                select_stmt = self._make_select_stmt(
                    select_stmt=select_stmt,
                    query=query.get_field_query(field_name),
                    mapper=get_related_mapper(
                        mapper=mapper,
                        field_name=field_name,
                    ),
                    label_prefix=field_name
                )

        return select_stmt

    async def _select(
        self,
        query: Union[Type[SelectQuery], SelectQuery],
        mapper: DeclarativeMeta
    ) -> List[BaseQuery]:
        select_stmt = self.make_select_stmt(
            query=query,
            mapper=mapper
        )

        scalar_result = await self._session.execute(select_stmt)

        return query.from_selected_rows(rows=scalar_result.all())



from sqlalchemy.sql import Select
from query.pagination_query import PaginationQuery
from .base import BaseProvider


class PaginationProvider(BaseProvider):
    def bind_pagination(
        self,
        query: PaginationQuery,
        select_stmt: Select
    ) -> Select:
        return select_stmt.limit(query.limit).offset(query.offset)



from __future__ import annotations
from typing import Any, Dict
from copy import deepcopy
from utils import cls_or_ins
from .base import BaseQuery
from .from_row import FromRowQuery
from .join_query import JoinQuery


class UpdateQuery(FromRowQuery, JoinQuery, BaseQuery):
    _values: Dict[str, Any] = dict()

    @classmethod
    def get_values(cls) -> Dict[str, Any]:
        return deepcopy(cls._values)

    @cls_or_ins
    def set_values(cls_or_ins, **kwargs) -> BaseQuery:
        self = cls_or_ins
        if cls_or_ins.is_class():
            self = cls_or_ins()

        self._set_values(self._values, **kwargs)

        return self



from __future__ import annotations
from typing import List, Any, Dict
from sqlalchemy.engine.row import Row
from sqlalchemy.orm import DeclarativeMeta
from .base import BaseQuery


FIELD_NAME_SEPARATOR = '__'


class FromRowQuery(BaseQuery):
    @classmethod
    def from_returning_mapper(
        cls,
        mapper: DeclarativeMeta
    ) -> BaseQuery:
        type_hints = cls.get_type_hints()
        query = cls()
        for field_name in type_hints.keys():
            setattr(query, field_name, getattr(mapper, field_name, None))

        return query

    @classmethod
    def from_returning_mappers(
        cls,
        mappers: List[DeclarativeMeta]
    ) -> List[BaseQuery]:
        queries = list()
        for mapper in mappers:
            query = cls.from_returning_mapper(mapper=mapper)
            queries.append(query)

        return queries

    @classmethod
    def _from_mapping(
        cls,
        mapping: Dict[str, Any]
    ) -> BaseQuery:
        query = cls()

        nested_mappings: Dict[str, Dict[str, Any]] = dict()

        for field, value in mapping.items():
            field_name, *deeper = field.split(FIELD_NAME_SEPARATOR)
            if not deeper:
                setattr(query, field_name, value)
                continue

            if field_name in nested_mappings:
                nested_mappings[field_name].update({
                    FIELD_NAME_SEPARATOR.join(deeper): value
                })
            else:
                nested_mappings[field_name] = {
                    FIELD_NAME_SEPARATOR.join(deeper): value
                }

        if not nested_mappings:
            return query

        for field_name, mapping in nested_mappings.items():
            nested_query = cls.get_field_query(field_name=field_name)
            nested_query = nested_query._from_mapping(mapping=mapping)
            setattr(query, field_name, nested_query)

        return query

    @classmethod
    def from_selected_row(
        cls,
        row: Row
    ) -> BaseQuery:
        return cls._from_mapping(mapping=dict(row))

    @classmethod
    def from_selected_rows(
        cls,
        rows: List[Row]
    ) -> List[BaseQuery]:
        queries = list()
        for row in rows:
            queries.append(cls.from_selected_row(row=row))

        return queries


a = {
    'id': 1,
    'name': 'some_name',
    'device_type__id': 'some_id',
    'device_type__name': 'some_name',
    'device_type__meter_inline__id': 3,
}

